@@ -3,6 +3,7 @@ import json
 import time
 import datetime
 import random
+import jwt
 
 
 class SearchAdsAPI:
@@ -10,8 +11,11 @@ class SearchAdsAPI:
                  org_id,
                  pem,
                  key,
+                 client_id=None,
+                 team_id=None,
+                 key_id=None,
                  certificates_dir_path="certs/",
-                 api_version="v3",
+                 api_version="v4",
                  session=None,
                  verbose=False):
         """
@@ -24,6 +28,60 @@ class SearchAdsAPI:
         self.path = certificates_dir_path
         self.verbose = verbose
         self.api_version = api_version
+        self.client_id = client_id
+        self.team_id = team_id
+        self.key_id = key_id
+
+    def get_access_token_from_client_secret(self,key):
+        audience = 'https://appleid.apple.com'
+        alg = 'ES256'
+        # Define issue timestamp.
+        issued_at_timestamp = int(datetime.datetime.utcnow().timestamp())
+        # Define expiration timestamp. May not exceed 180 days from issue timestamp.
+        expiration_timestamp = issued_at_timestamp + 86400*180
+
+        # Define JWT headers.
+        headers = dict()
+        headers['alg'] = alg
+        headers['kid'] = self.key_id
+
+        # Define JWT payload.
+        payload = dict()
+        payload['sub'] = self.client_id
+        payload['aud'] = audience
+        payload['iat'] = issued_at_timestamp
+        payload['exp'] = expiration_timestamp
+        payload['iss'] = self.team_id
+
+        # Path to signed private key.
+        KEY_FILE = key
+
+        with open(KEY_FILE, 'r') as key_file:
+            key = ''.join(key_file.readlines())
+
+        client_secret = jwt.encode(
+            payload=payload,
+            headers=headers,
+            algorithm=alg,
+            key=key
+        )
+        # with open(f'{KEY_FILE}.txt', 'w') as output:
+        #     output.write(client_secret.decode("utf-8"))
+        result = requests.post(f"""https://appleid.apple.com/auth/oauth2/token""",
+                      data={"grant_type": "client_credentials",
+                            "client_id": f"{self.client_id}",
+                            "client_secret": f"{client_secret}",
+                            "scope": "searchadsorg"},
+                      headers={
+                          "Host": "appleid.apple.com",
+                          "Content-Type": "application/x-www-form-urlencoded"
+                      })
+        
+        # return client_secret
+        access_token = result.json()["access_token"]
+        print(access_token)
+        return access_token
+        
 
     # API Function
     def api_call(self,
@@ -31,57 +89,66 @@ class SearchAdsAPI:
                  headers={},
                  json_data={},
                  params={},
-                 method="GET", 
+                 method="GET",
                  limit=1000,
                  offset=0):
         """
         Generic API call function
         """
         url = "https://api.searchads.apple.com/api/{}"
+        # choose between using a proxy or the public ip
         if self.session is None:
             caller = requests
         else:
             caller = self.session
+        # find the certicates path
         pem = self.path + self.pem
         key = self.path + self.key
+        
         kwargs = {
-            "cert": (
-                pem,
-                key
-            ),
             "headers": headers,
         }
+        # if v3 is being used
+        if self.client_id is None:
+            kwargs["cert"] = (
+                pem,
+                key
+            )
         if json_data:
             kwargs['json'] = json_data
         kwargs["params"] = dict()
+        # add the limit if it applies
         if limit:
             kwargs['params']['limit'] = limit
+        # add the offset
         if offset:
             kwargs['params']['offset'] = offset
+        # add the org_id header in v3
         if self.org_id:
-            kwargs['headers']["Authorization"] = "orgId={org_id}".format(
-                org_id=self.org_id)
+            kwargs['headers']["Authorization"] = f"orgId={self.org_id}"
+        # only if using Search Ads API v4
+        if self.key_id is not None:
+            access_token = self.get_access_token_from_client_secret(key)
+            kwargs['headers']["Authorization"] = f"Bearer {access_token}"
+            kwargs['headers']["X-AP-Context"] = f"orgId={self.org_id}"
         kwargs["params"].update(params)
         api_endpoint = "{}/{}".format(self.api_version, api_endpoint)
         url = url.format(api_endpoint)
-        try:
-            if method == "get" or method == "GET":
-                req = caller.get(url, **kwargs)
-            elif method == "post" or method == "POST":
-                req = caller.post(url, **kwargs)
-            elif method == "put" or method == "PUT":
-                req = caller.put(url, **kwargs)
-            elif method == "delete" or method == "DELETE":
-                req = caller.delete(url, **kwargs)
-            if self.verbose:
-                print(req.url)
-                print(req.text)
-            return req.json()
-        except Exception as e:
-            print(str(e), url)
-            return None
+        if method == "get" or method == "GET":
+            req = caller.get(url, **kwargs)
+        elif method == "post" or method == "POST":
+            req = caller.post(url, **kwargs)
+        elif method == "put" or method == "PUT":
+            req = caller.put(url, **kwargs)
+        elif method == "delete" or method == "DELETE":
+            req = caller.delete(url, **kwargs)
+        if self.verbose:
+            print(req.url)
+            print(req.text)
+        return req.json()
 
     # Campaign Methods
+
     def create_campaign(self,
                         app_id,
                         countries,
@@ -1456,9 +1523,9 @@ class SearchAdsAPI:
                 grandTotals.extend(
                     res["data"]["reportingDataResponse"]["grandTotals"])
             res_len = len(row)
-            
+
             offset = len(row)
-                # print(limit)
+            # print(limit)
             if res_len == limit or res_len >= res["pagination"]["totalResults"]:
                 break
         if return_grand_totals:
